from ordeq.framework import (
    IO,
    Input,
    InputHook,
    IOException,
    Node,
    NodeHook,
    NodeNotFound,
    Output,
    OutputHook,
    RunHook,
    node,
    run,
)

__all__ = (
    "IO",
    "IOException",
    "Input",
    "InputHook",
    "Node",
    "NodeHook",
    "NodeNotFound",
    "Output",
    "OutputHook",
    "RunHook",
    "node",
    "run",
)



from types import ModuleType

from ordeq.framework.io import IO, Input, Output


class CatalogError(Exception): ...


def check_catalogs_are_consistent(
    a: ModuleType, b: ModuleType, *others: ModuleType
) -> None:
    """Utility method to checks if two (or more) catalogs are consistent,
    i.e. if they define the same keys.

    Args:
        a: First catalog to compare.
        b: Second catalog to compare.
        *others: Additional catalogs to compare.

    Raises:
        CatalogError: If the catalogs are inconsistent,
            i.e. if they define different keys.
    """

    ios = []  # for each catalog, the names (keys) of the IOs it defines
    for catalog in [a, b, *others]:
        catalog_ios = set()
        for key, value in vars(catalog).items():
            if isinstance(value, (IO, Input, Output)):
                catalog_ios.add(key)
        ios.append(catalog_ios)

    if not all(s == ios[0] for s in ios[1:]):
        raise CatalogError("Catalogs are inconsistent.")



import logging
from collections.abc import Callable, Sequence
from itertools import chain
from types import ModuleType
from typing import Literal, TypeVar

from ordeq.framework._resolve import _resolve_runnables_to_nodes
from ordeq.framework.graph import NodeGraph
from ordeq.framework.hook import NodeHook, RunHook
from ordeq.framework.io import Input, Output, _InputCache
from ordeq.framework.nodes import Node

logger = logging.getLogger(__name__)

T = TypeVar("T")

DataStoreType = dict[Input[T] | Output[T], T]
# The save mode determines which outputs are saved. When set to:
# - 'all', all outputs are saved, including those of intermediate nodes.
# - 'sinks', only outputs of sink nodes are saved, i.e. those w/o successors.
# - 'none', to dry-run and save no outputs
# Future extension:
# - 'last', which saves the output of the last node for which no error
# occurred. This can be useful for debugging.
SaveMode = Literal["all", "sinks", "none"]


def _save_outputs(
    node: Node, values: Sequence[T], save: bool = True
) -> dict[Input[T] | Output[T], T]:
    computed: dict[Input[T] | Output[T], T] = {}
    for output_dataset, data in zip(node.outputs, values, strict=False):
        computed[output_dataset] = data

        # TODO: this can be handled in the `save_wrapper`
        if save:
            output_dataset.save(data)

    return computed


def _run_node(
    node: Node, *, hooks: Sequence[NodeHook] = (), save: bool = True
) -> DataStoreType:
    node.validate()

    for node_hook in hooks:
        node_hook.before_node_run(node)

    args = [input_dataset.load() for input_dataset in node.inputs]

    # persisting loaded data
    for node_input, data in zip(node.inputs, args, strict=True):
        if isinstance(node_input, _InputCache):
            node_input.persist(data)

    logger.info("Running node %s", node)

    try:
        values = node.func(*args)
    except Exception as exc:
        for node_hook in hooks:
            node_hook.on_node_call_error(node, exc)
        raise exc

    if len(node.outputs) == 0:
        values = ()
    elif len(node.outputs) == 1:
        values = (values,)
    else:
        values = tuple(values)

    computed = _save_outputs(node, values, save=save)

    # persisting computed data only if outputs are loaded again later
    for node_output in node.outputs:
        if isinstance(node_output, _InputCache):
            node_output.persist(computed[node_output])  # ty: ignore[call-non-callable]

    for node_hook in hooks:
        node_hook.after_node_run(node)

    return computed


def _run_graph(
    graph: NodeGraph,
    *,
    hooks: Sequence[NodeHook] = (),
    save: SaveMode = "all",
    io: dict[Input[T] | Output[T], Input[T] | Output[T]] | None = None,
) -> DataStoreType:
    """Runs nodes in a graph topologically, ensuring IOs are loaded only once.

    Args:
        graph: node graph to run
        hooks: hooks to apply
        hooks: hooks to apply
        save: 'all' | 'sinks' | 'none'.
            If 'sinks', only saves the outputs of sink nodes in the graph.
        io: mapping of IO objects to their replacements

    Returns:
        a dict mapping each IO to the computed data

    """

    patched_nodes: dict[Node, Node] = {}
    for node in graph.nodes:
        patched_nodes[node] = _patch_io(node, io or {})

    data_store: dict = {}  # For each IO, the loaded data

    # TODO: Create _Patch wrapper for IO?
    for node in graph.topological_ordering:
        if (save == "sinks" and node in graph.sink_nodes) or save == "all":
            save_node = True
        else:
            save_node = False

        computed = _run_node(patched_nodes[node], hooks=hooks, save=save_node)
        data_store.update(computed)

    reverse_io = {v: k for k, v in (io or {}).items()}
    patched_data_store = {}
    for k, v in data_store.items():
        patched_data_store[reverse_io.get(k, k)] = v

    # unpersist IO objects
    for gnode in graph.nodes:
        io_objs = chain(gnode.inputs, gnode.outputs)
        for io_obj in io_objs:
            if isinstance(io_obj, _InputCache):
                io_obj.unpersist()

    return patched_data_store


def _patch_io(
    node: Node, io: dict[Input[T] | Output[T], Input[T] | Output[T]]
) -> Node:
    """Patches the inputs and outputs of a node with the provided IO mapping.

    Args:
        node: the original node
        io: mapping of Input/Output objects to their replacements

    Returns:
        the node with patched inputs and outputs

    """

    return node._replace(
        inputs=[io.get(ip, ip) for ip in node.inputs],  # type: ignore[misc]
        outputs=[io.get(op, op) for op in node.outputs],  # type: ignore[misc]
    )


def run(
    *runnables: ModuleType | Callable | str,
    hooks: Sequence[NodeHook | RunHook] = (),
    save: SaveMode = "all",
    verbose: bool = False,
    io: dict[Input[T] | Output[T], Input[T] | Output[T]] | None = None,
) -> DataStoreType:
    """Runs nodes in topological order.

    Args:
        runnables: the nodes to run, or modules or packages containing nodes
        hooks: hooks to apply
        save: 'all' | 'sinks'. If 'sinks', only saves the sink outputs
        verbose: whether to print the node graph
        io: mapping of IO objects to their replacements

    Returns:
        a dict mapping each IO to the computed data

    """

    nodes = _resolve_runnables_to_nodes(*runnables)
    graph = NodeGraph.from_nodes(nodes)

    if verbose:
        print(graph)

    node_hooks: list[NodeHook] = [h for h in hooks if isinstance(h, NodeHook)]
    run_hooks: list[RunHook] = [h for h in hooks if isinstance(h, RunHook)]

    for run_hook in run_hooks:
        run_hook.before_run(graph)

    result = _run_graph(graph, hooks=node_hooks, save=save, io=io)

    for run_hook in run_hooks:
        run_hook.after_run(graph, result)

    return result



import importlib
import logging
from collections.abc import Callable, Sequence
from dataclasses import dataclass, field, replace
from functools import cached_property, wraps
from inspect import Signature, signature
from typing import Any, Generic, ParamSpec, TypeVar, overload

from ordeq.framework._registry import NODE_REGISTRY
from ordeq.framework.io import Input, Output

logger = logging.getLogger(__name__)

T = TypeVar("T")
FuncParams = ParamSpec("FuncParams")
FuncReturns = TypeVar("FuncReturns")


@dataclass(frozen=True)
class Node(Generic[FuncParams, FuncReturns]):
    func: Callable[FuncParams, FuncReturns]
    inputs: tuple[Input, ...]
    outputs: tuple[Output, ...]
    tags: list[str] | dict[str, Any] = field(default_factory=list, hash=False)

    def __post_init__(self):
        """Nodes always have to be hashable"""
        _raise_if_not_hashable(self)
        if self.inputs:
            _raise_for_invalid_inputs(self)
        if self.outputs:
            _raise_for_invalid_outputs(self)

    def validate(self):
        """These checks are performed before the node is run."""
        _raise_for_invalid_inputs(self)
        _raise_for_invalid_outputs(self)

    @cached_property
    def name(self) -> str:
        full_name = self.func.__name__
        if hasattr(self.func, "__module__"):
            module = str(self.func.__module__)
            if module != "__main__":
                full_name = module + ":" + full_name
        return full_name

    def __repr__(self) -> str:
        attributes = {"name": self.name}

        inputs = getattr(self, "inputs", None)
        if inputs:
            input_str = ", ".join(repr(i) for i in inputs)
            attributes["inputs"] = f"[{input_str}]"

        outputs = getattr(self, "outputs", None)
        if outputs:
            output_str = ", ".join(repr(o) for o in outputs)
            attributes["outputs"] = f"[{output_str}]"

        if self.tags:
            attributes["tags"] = repr(self.tags)

        attributes_str = ", ".join(f"{k}={v}" for k, v in attributes.items())
        return f"Node({attributes_str})"

    def _replace(
        self,
        inputs: Sequence[Input] | Input,
        outputs: Sequence[Output] | Output,
    ) -> "Node[FuncParams, FuncReturns]":
        return replace(
            self,
            inputs=_sequence_to_tuple(inputs),
            outputs=_sequence_to_tuple(outputs),
        )


def _raise_for_invalid_inputs(n: Node) -> None:
    """Raises a ValueError if the number of inputs is incompatible with
    node arguments.

    Args:
        n: a Node

    Raises:
        ValueError: if the number of inputs is incompatible with the number of
            node arguments.
    """

    func = n.func
    sign = signature(func)
    try:
        sign.bind(*n.inputs)
    except TypeError as e:
        raise ValueError(
            f"Node inputs invalid for function arguments: "
            f"Node(name={n.name},...)"
        ) from e


def _raise_for_invalid_outputs(n: Node) -> None:
    """Raises a ValueError if the number of outputs is incompatible with
    node arguments.

    Args:
        n: a Node

    Raises:
        ValueError: if the number of outputs is incompatible with the number of
            node arguments.
    """
    func = n.func
    sign = signature(func)
    returns = sign.return_annotation
    if returns == Signature.empty:
        return

    # deal with `from __future__ import annotations`
    if isinstance(returns, str):
        try:
            mod = importlib.import_module(func.__module__)
            returns = eval(returns, mod.__dict__)  # noqa: S307
        except (NameError, ImportError):
            return

    # any return type is valid for a single output
    if len(n.outputs) == 1:
        return

    # A type annotation was provided
    if returns is None:
        return_types = []
    elif hasattr(returns, "__origin__") and returns.__origin__ is tuple:
        # tuple[pd.DataFrame, list[str]] => 2
        return_types = returns.__args__
    else:
        return_types = [returns]

    if len(return_types) != len(n.outputs):
        raise ValueError(
            "Node outputs invalid for return annotation: "
            f"Node(name={n.name},...). "
            f"Node has {len(n.outputs)} output(s), but the return type "
            f"annotation expects {len(return_types)} value(s)."
        )


def _raise_if_not_hashable(n: Node) -> None:
    """Raises a ValueError if a node is not hashable.

    Args:
        n: a Node

    Raises:
        ValueError: if the node is not hashable
    """

    try:
        hash(n)
    except TypeError as e:
        raise ValueError(
            f"Node is not hashable: Node(name={n.name}, ...)"
        ) from e


def _sequence_to_tuple(obj: Sequence[T] | T | None) -> tuple[T, ...]:
    if obj is None:
        return ()
    if isinstance(obj, Sequence):
        return tuple(obj)
    return (obj,)  # ty: ignore[invalid-return-type]


def _create_node(
    func: Callable[FuncParams, FuncReturns],
    inputs: Sequence[Input] | Input | None = None,
    outputs: Sequence[Output] | Output | None = None,
    tags: list[str] | dict[str, Any] | None = None,
) -> Node[FuncParams, FuncReturns]:
    n = Node(
        func,
        _sequence_to_tuple(inputs),
        _sequence_to_tuple(outputs),
        [] if tags is None else tags,
    )
    NODE_REGISTRY.set(func, n)
    return n


@overload
def node(
    func: Callable[FuncParams, FuncReturns],
    *,
    inputs: Sequence[Input] | Input | None = None,
    outputs: Sequence[Output] | Output | None = None,
    tags: list[str] | dict[str, Any] | None = None,
) -> Callable[FuncParams, FuncReturns]: ...


@overload
def node(
    *,
    inputs: Sequence[Input] | Input | None = None,
    outputs: Sequence[Output] | Output | None = None,
    tags: list[str] | dict[str, Any] | None = None,
) -> Callable[
    [Callable[FuncParams, FuncReturns]], Callable[FuncParams, FuncReturns]
]: ...


def node(
    func: Callable[FuncParams, FuncReturns] | None = None,
    *,
    inputs: Sequence[Input] | Input | None = None,
    outputs: Sequence[Output] | Output | None = None,
    tags: list[str] | dict[str, Any] | None = None,
) -> (
    Callable[
        [Callable[FuncParams, FuncReturns]], Callable[FuncParams, FuncReturns]
    ]
    | Callable[FuncParams, FuncReturns]
):
    """Decorator that creates a node from a function. When a node is run,
    the inputs are loaded and passed to the function. The returned values
    are saved to the outputs.

    Example:

    ```python
    >>> from pyspark.sql import DataFrame
    >>> @node(
    ...     inputs=CSV(path="path/to.csv"),
    ...     outputs=Table(table="db.table")
    ... )
    ... def transformation(csv: DataFrame) -> DataFrame:
    ...     return csv.select("someColumn")

    ```

    Nodes can also take a variable number of inputs:

    ```python
    >>> @node(
    ...     inputs=[
    ...         CSV(path="path/to/fst.csv"),
    ...         # ...
    ...         CSV(path="path/to/nth.csv")
    ...     ],
    ...     outputs=Table(table="db.all_appended")
    ... )
    ... def append_dfs(*args: DataFrame) -> DataFrame:
    ...     df = args[0]
    ...     for arg in args[1:]:
    ...         df = df.unionByName(arg)
    ...     return df

    ```

    Node can also be created from existing functions:

    ```python
    >>> def remove_header(data: list[str]) -> list[str]:
    ...     return data[1:]
    >>> fst = node(remove_header, inputs=CSV(path="path/to/fst.csv"), ...)
    >>> snd = node(remove_header, inputs=CSV(path="path/to/snd.csv"), ...)
    >>> ...

    ```

    You can assign tags to a node, which can be used for filtering or grouping
    nodes later:

    ```python
    >>> @node(inputs=..., outputs=..., tags=["tag1", "tag2"])
    ... def func(...): -> ...

    ```

    Args:
        func: function of the node
        inputs: sequence of inputs
        outputs: sequence of outputs
        tags: tags to assign to the node

    Returns:
        a node

    """

    if func is None:
        # we are called as @node(inputs=...
        def wrapped(
            f: Callable[FuncParams, FuncReturns],
        ) -> Callable[FuncParams, FuncReturns]:
            @wraps(f)
            def inner(*args, **kwargs):
                # Purpose of this inner is to create a new function from `f`
                return f(*args, **kwargs)

            _create_node(inner, inputs, outputs, tags)
            return inner

        return wrapped

    # else: we are called as node(func, inputs=...)

    @wraps(func)
    def wrapper(*args, **kwargs):
        # The purpose of this wrapper is to create a new function from `func`
        return func(*args, **kwargs)

    _create_node(wrapper, inputs, outputs, tags)
    return wrapper


class NodeNotFound(Exception):
    """Exception raised when a Node is not found in the registry."""


def get_node(
    func: Callable[FuncParams, FuncReturns],
) -> Node[FuncParams, FuncReturns]:
    """Retrieve a Node from the node registry based on a function.

    Args:
        func: the function for which to retrieve the Node

    Returns:
        the Node

    Raises:
        NodeNotFound: if the node is not found in the registry.
    """

    try:
        return NODE_REGISTRY.get(func)
    except KeyError:
        raise NodeNotFound(
            f"Node '{func.__name__}' not found in the registry"
        ) from None



"""Resolve packages and modules to nodes and IOs."""

import importlib
import pkgutil
from collections.abc import Callable, Generator, Hashable, Iterable
from types import ModuleType

from ordeq.framework._registry import NODE_REGISTRY
from ordeq.framework.io import IO, Input, Output
from ordeq.framework.nodes import Node, get_node


def _is_module(obj: object) -> bool:
    return isinstance(obj, ModuleType)


def _is_package(module: ModuleType) -> bool:
    return hasattr(module, "__path__")


def _is_io(obj: object) -> bool:
    return isinstance(obj, (IO, Input, Output))


def _is_node(obj: object) -> bool:
    return isinstance(obj, Hashable) and obj in NODE_REGISTRY


def _resolve_string_to_module(name: str) -> ModuleType:
    return importlib.import_module(name)


def _resolve_packages_to_modules(
    modules: Iterable[tuple[str, ModuleType]],
) -> Generator[tuple[str, ModuleType], None, None]:
    for name, module in modules:
        yield name, module
        if _is_package(module):
            submodules = (
                importlib.import_module(f".{name}", package=module.__name__)
                for _, name, _ in pkgutil.iter_modules(module.__path__)
            )
            yield from _resolve_packages_to_modules(
                (mod.__name__, mod) for mod in submodules
            )


def _resolve_runnables_to_modules(
    *runnables: str | ModuleType,
) -> Generator[tuple[str, ModuleType], None, None]:
    modules = {}
    for runnable in runnables:
        if isinstance(runnable, ModuleType):
            modules[runnable.__name__] = runnable
        elif isinstance(runnable, str):
            mod = _resolve_string_to_module(runnable)
            modules[mod.__name__] = mod
        else:
            raise TypeError(
                f"{runnable} is not something we can run. "
                f"Expected a module or a string, got {type(runnable)}"
            )

    # Then, for each module or package, if it's a package, resolve to all its
    # submodules recursively
    return _resolve_packages_to_modules(modules.items())


def _resolve_module_to_nodes(module: ModuleType) -> set[Node]:
    """Gathers all nodes defined in a module.

    Args:
        module: the module to gather nodes from

    Returns:
        the nodes defined in the module

    """

    return {get_node(obj) for obj in vars(module).values() if _is_node(obj)}


def _resolve_module_to_ios(
    module: ModuleType,
) -> dict[str, IO | Input | Output]:
    """Find all `IO` objects defined in the provided module

    Args:
        module: the Python module object

    Returns:
        a dict of `IO` objects with their variable name as key
    """
    return {name: obj for name, obj in vars(module).items() if _is_io(obj)}


def _resolve_runnables_to_nodes_and_modules(
    *runnables: str | ModuleType | Callable,
) -> tuple[set[Node], set[ModuleType]]:
    """Collects nodes and modules from the provided runnables.

    Args:
        runnables: modules, packages or callables to gather nodes from

    Returns:
        the nodes and modules collected from the runnables

    Raises:
        TypeError: if a runnable is not a module and not a node
    """
    modules_and_strs = []
    nodes = set()
    for runnable in runnables:
        if isinstance(runnable, (ModuleType, str)):
            modules_and_strs.append(runnable)
        elif callable(runnable):
            nodes.add(get_node(runnable))
        else:
            raise TypeError(
                f"{runnable} is not something we can run. "
                f"Expected a module or a node, got {type(runnable)}"
            )

    modules = set(
        dict(_resolve_runnables_to_modules(*modules_and_strs)).values()
    )
    return nodes, modules


def _resolve_runnables_to_nodes(
    *runnables: ModuleType | Callable | str,
) -> set[Node]:
    """Collects nodes from the provided runnables.

    Args:
        runnables: modules, packages or callables to gather nodes from

    Returns:
        the nodes collected from the runnables

    """

    nodes, modules = _resolve_runnables_to_nodes_and_modules(*runnables)
    for module in modules:
        nodes.update(_resolve_module_to_nodes(module))

    return nodes


def _gather_nodes_from_registry() -> set[Node]:
    """Find all `Node` objects defined in the provided module

    Returns:
        a set of `Node` objects
    """
    return set(NODE_REGISTRY._data.values())  # noqa: SLF001


def _check_missing_ios(
    nodes: set[Node], ios: dict[str, IO | Input | Output]
) -> None:
    missing_ios: set[IO | Input | Output] = set()
    for node in nodes:
        for inp in node.inputs:
            if inp not in ios.values():
                missing_ios.add(inp)
        for out in node.outputs:
            if out not in ios.values():
                missing_ios.add(out)

    if missing_ios:
        raise ValueError(
            f"The following IOs are used by nodes but not defined: "
            f"{missing_ios}. Please include the module defining them in "
            f"the runnables."
        )


def _resolve_runnables_to_nodes_and_ios(
    *runnables: str | ModuleType | Callable,
) -> tuple[set[Node], dict[str, IO | Input | Output]]:
    """Collects nodes and IOs from the provided runnables.

    Args:
        runnables: package names, modules, or callables to gather nodes and IOs
            from

    Returns:
        a tuple of nodes and IOs collected from the runnables
    """

    ios = {}
    nodes, modules = _resolve_runnables_to_nodes_and_modules(*runnables)
    for node in nodes:
        mod = _resolve_string_to_module(node.func.__module__)
        ios.update(_resolve_module_to_ios(mod))

    for module in modules:
        nodes.update(_resolve_module_to_nodes(module))
        ios.update(_resolve_module_to_ios(module))

    return nodes, ios



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Protocol,
    TypeAlias,
    TypeVar,
    runtime_checkable,
)

if TYPE_CHECKING:
    from ordeq.framework.graph import NodeGraph
    from ordeq.framework.io import IO, Input, Output
    from ordeq.framework.nodes import Node

T = TypeVar("T")


@runtime_checkable
class InputHook(Protocol[T]):
    """Hook used to inject custom logic that will be executed before and after
    an input is loaded."""

    def before_input_load(self, io: Input[T]) -> None:
        """Hook that is executed before an input is loaded.

        Args:
            io: the input object.
        """

        return

    def after_input_load(self, io: Input[T], data: T) -> None:
        """Hook that is executed after an input is loaded.

        Args:
            io: the input object.
            data: the loaded data.
        """

        return


@runtime_checkable
class OutputHook(Protocol[T]):
    """Hook used to inject custom logic that will be executed before and after
    an output is saved."""

    def before_output_save(self, io: Output[T], data: T) -> None:
        """Hook that is executed before an output is saved.

        Args:
            io: the input object.
            data: the data to be saved.
        """

        return

    def after_output_save(self, io: Output[T], data: T) -> None:
        """Hook that is executed after an output is saved.

        Args:
            io: the input object.
            data: the data that has been saved.
        """

        return


@runtime_checkable
class NodeHook(Protocol):
    """Hook used to inject custom logic that will be executed when a node is
    run."""

    def before_node_run(self, node: Node) -> None:
        return None

    def on_node_call_error(self, node: Node, error: Exception) -> None:
        """Triggered when an exception is raised when calling the node.

        Args:
            node: the node object
            error: the error
        """

        return

    def after_node_run(self, node: Node) -> None:
        return None


@runtime_checkable
class RunHook(Protocol):
    """Hook used to inject custom logic that will be executed when a graph is
    run."""

    def before_run(self, graph: NodeGraph) -> None:
        """Triggered before the graph is run.

        Args:
            graph: the graph object
        """

        return

    def after_run(
        self, graph: NodeGraph, data: dict[Input | Output | IO, Any]
    ) -> None:
        """Triggered after the graph is run.

        Args:
            graph: the graph object
            data: mapping of all inputs and outputs to their loaded/saved data
        """
        return


Hook: TypeAlias = InputHook | OutputHook | NodeHook | RunHook
